import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter_hooks/flutter_hooks.dart';
import 'package:flutter_query/flutter_query.dart';
import 'package:flutter_query/src/query.dart';
import 'package:provider/provider.dart';

typedef Future<T> QueryFunction<T>();
typedef Future PageableQueryFunction<T>(int page);
typedef dynamic GetFetchMoreFunction(latestPage);

AsyncSnapshot<T> useQuery<T>(String key, QueryFunction<T> fn) {
  return use(_QueryHook<T>(key, fn));
}

InfiniteQueryContext<T> useInfiniteQuery<T>(
  String key,
  PageableQueryFunction<T> fn, {
  GetFetchMoreFunction getFetchMore,
}) {
  return use(_InfiniteQueryHook<T>(key, fn, getFetchMore));
}

class _QueryHook<T> extends Hook<AsyncSnapshot<T>> {
  const _QueryHook(this.key, this.fn);

  final String key;
  final QueryFunction<T> fn;

  @override
  _QueryStateHook<T> createState() => _QueryStateHook();
}

class _QueryStateHook<T> extends HookState<AsyncSnapshot<T>, _QueryHook<T>> {
  AsyncSnapshot<T> _snapshot;
  QueryCache _cache;
  StreamSubscription<AsyncSnapshot<T>> _subscription;

  @override
  void initHook() {
    _cache = Provider.of<QueryCache>(context, listen: false);
    final key = hook.key;
    _cache.registerQuery<T>(key, QueryConfig(fn: hook.fn));
    final watcher = (AsyncSnapshot<T> snapshot) {
      setState(() {
        _snapshot = snapshot;
      });
    };
    _subscription = _cache.watchQuery(key, watcher);
    super.initHook();
  }

  @override
  void dispose() {
    _subscription.cancel();
    super.dispose();
  }

  @override
  AsyncSnapshot<T> build(BuildContext context) {
    return _snapshot;
  }
}

class InfiniteQueryContext<T> {
  AsyncSnapshot<T> snapshot;
  Function fetchMore;
  bool hasMorePages = true;
}

class _InfiniteQueryHook<T> extends Hook<InfiniteQueryContext<T>> {
  const _InfiniteQueryHook(this.key, this.fn, this.getFetchMore);

  final String key;
  final PageableQueryFunction<T> fn;
  final Function getFetchMore;

  @override
  HookState<InfiniteQueryContext<T>, Hook<InfiniteQueryContext<T>>>
      createState() => _InfiniteQueryStateHook();
}

class _InfiniteQueryStateHook<T>
    extends HookState<InfiniteQueryContext<T>, _InfiniteQueryHook<T>> {
  InfiniteQueryContext<T> _context;
  QueryCache _cache;
  StreamSubscription<AsyncSnapshot<T>> _subscription;

  @override
  void initHook() {
    _context = InfiniteQueryContext<T>();

    try {
      _cache = Provider.of<QueryCache>(context, listen: false);
      final key = hook.key;
      _cache.registerQuery(
        key,
        QueryConfig(
          fn: hook.fn,
          getFetchMore: hook.getFetchMore ?? (lastCursor) => lastCursor + 1,
          isInfinite: true,
        ),
      );
      _context.fetchMore = () {
        _cache.invalidateQuery(key);
      };
      final watcher = (snapshot) {
        setState(() {
          _context.snapshot = snapshot;
        });
      };
      _subscription = _cache.watchQuery(key, watcher);
    } catch (e) {
      throw Exception(
        'Cache not found! You either need to use the QueryCacheProvider or provide a query cache yourself.',
      );
    }
    super.initHook();
  }

  @override
  void dispose() {
    _subscription.cancel();
    super.dispose();
  }

  @override
  InfiniteQueryContext<T> build(BuildContext context) {
    return _context;
  }
}
